
# -- import packages: ---------------------------------------------------------
import ABCParse
import adata_query
import anndata
import scanpy as sc


# -- set typing: --------------------------------------------------------------
from typing import Optional, Union


# -- Operational class: -------------------------------------------------------
class SCVerseNeighbors(ABCParse.ABCParse):
    def __init__(
        self,
        distance_key: str = "distances",
        connectivities_key: str = "connectivities",
        params_key: str = "neighbors",
        silent: bool = False,
        *args,
        **kwargs,
    ):
        self.__parse__(locals())

        self._neighbor_computation_performed = False

    @property
    def _SCANPY_NEIGHBORS_KWARGS(self):
        return ABCParse.function_kwargs(sc.pp.neighbors, self._PARAMS)

    @property
    def _HAS_DISTANCES(self):
        return self._distance_key in self._adata.obsp

    @property
    def _HAS_CONNECTIVITIES(self):
        return self._connectivities_key in self._adata.obsp

    @property
    def _HAS_NN_PARAMS(self):
        return self._params_key in self._adata.uns

    def _probable_fetch(self, key):
        return adata_query.fetch(self._adata, key)

    @property
    def distances(self):
        if self._HAS_DISTANCES:
            return self._probable_fetch(self._distance_key)

    @property
    def connectivities(self):
        if self._HAS_CONNECTIVITIES:
            return self._probable_fetch(self._connectivities_key)

    @property
    def params(self):
        if self._HAS_NN_PARAMS:
            return self._probable_fetch(self._params_key)

    @property
    def properties(self):
        return {
            attr[1:]: getattr(self, attr) for attr in self.__dir__() if "_HAS_" in attr
        }

    @property
    def neighbors_precomputed(self):
        return all(self.properties.values())

    def _intake(self, adata):
        """"""

        self._preexisting = {}

        for key in ["obsp", "uns"]:
            if not hasattr(adata, key):
                self._preexisting[key] = None
            else:
                self._preexisting[key] = list(getattr(adata, key).keys())

    def _message(self, adata):
        for key, val in self._preexisting.items():
            added = [attr for attr in getattr(adata, key).keys() if not attr in val]
            for added_val in added:
                self._INFO(f"Added: adata.{key}['{added_val}']")
                
    def forward(self, adata: anndata.AnnData, **kwargs):
        """"""
        self.__update__(kwargs)
        
        self._intake(adata)

        if self.neighbors_precomputed and self._force:
            if not self._silent:
                self._INFO("Force recomputing neighbors")

        if not self.neighbors_precomputed or self._force:
            kw = self._SCANPY_NEIGHBORS_KWARGS.copy()
            kw.update({"adata": adata})

            sc.pp.neighbors(**kw)

            self._neighbor_computation_performed = True
        if not self._silent:
            self._message(adata)


    def __call__(
        self,
        adata: anndata.AnnData,
        n_neighbors: int = 15,
        n_pcs: Optional[int] = None,
        use_rep: Optional[str] = None,
        random_state: Optional[Union[int, None]] = 0,
        method: Optional = 'umap',
        metric: Optional = 'euclidean',
        distance_key: Optional[str] = "distances",
        connectivities_key: Optional[str] = "connectivities",
        params_key: Optional[str] = "neighbors",
        force: bool = False,
        silent: Optional[bool] = False,
        return_cls: Optional[bool] = False,
        *args,
        **kwargs,
    ):
        self.__update__(locals())
        
        self.forward(adata)


# -- API-facing function: -----------------------------------------------------
def scverse_neighbors(
    adata: anndata.AnnData,
    n_neighbors: int = 15,
    n_pcs: Optional[int] = None,
    use_rep: Optional[str] = None,
    random_state: Optional[Union[int, None]] = 0,
    method: Optional = 'umap',
    metric: Optional = 'euclidean',
    distance_key: Optional[str] = "distances",
    connectivities_key: Optional[str] = "connectivities",
    params_key: Optional[str] = "neighbors",
    force: Optional[bool] = False,
    silent: Optional[bool] = False,
    return_cls: Optional[bool] = False,
    *args,
    **kwargs,
):
    """Scanpy's ``sc.pp.neighbors`` with a few book-keeping functions added.
    
    Args:
        adata: anndata.AnnData
        
        n_neighbors (int): decsription. **Default** = 15
        
        distance_key (str): decsription. **Default** = "distances"
        
        connectivities_key (str): decsription. **Default** = "connectivities"
        
        params_key (str): decsription. **Default** = "neighbors"
        
        force (bool): decsription. **Default** = False
        
        return_cls (bool): Return the operator class for access to functional handles. **Default** = False
        
    Returns:
        None, updates ``adata``.
    """
        
    scv_neighbors = SCVerseNeighbors(
        distance_key=distance_key,
        connectivities_key=connectivities_key,
        params_key=params_key,
        silent=silent,
    )
    
    funcs = [sc.pp.neighbors, scv_neighbors.__call__]
    
    for func in funcs:
        kwargs.update(ABCParse.function_kwargs(func, locals()))
    
    scv_neighbors(**kwargs)
    
    if return_cls:
        return scv_neighbors




from neural_diffeqs import PotentialSDE
import torch


from . import base, mix_ins
from typing import Optional, Union, List

from ... import __version__


# -- lightning model: -----------------------------------
class LightningSDE_FixedPotential_RegularizedVelocityRatio(
    mix_ins.PotentialMixIn,
#     mix_ins.BaseForwardMixIn,
    mix_ins.RegularizedVelocityRatioMixIn,
    base.BaseLightningDiffEq,
):
    def __init__(
        self,
        latent_dim: int = 50,
        velocity_ratio_target: float = 1,
        velocity_ratio_enforce: float = 100,
        disable_velocity_ratio_backprop: bool = False,
        name: Optional[str] = None,
        mu_hidden: Union[List[int], int] = [2000, 2000],
        sigma_hidden: Union[List[int], int] = [800, 800],
        mu_activation: Union[str, List[str]] = 'LeakyReLU',
        sigma_activation: Union[str, List[str]] = 'LeakyReLU',
        mu_dropout: Union[float, List[float]] = 0.1,
        sigma_dropout: Union[float, List[float]] = 0.1,
        mu_bias: bool = True,
        sigma_bias: List[bool] = True,
        mu_output_bias: bool = True,
        sigma_output_bias: bool = True,
        mu_n_augment: int = 0,
        sigma_n_augment: int = 0,
        sde_type='ito',
        noise_type='general',
        brownian_dim=1,
        coef_drift: float = 1.0,
        coef_diffusion: float = 1.0,
        
        train_lr: float = 1e-4,
        train_optimizer=torch.optim.RMSprop,
        train_scheduler=torch.optim.lr_scheduler.StepLR,
        train_step_size: int = 10,
        dt: float = 0.1,
        adjoint = False,
        backend = "auto",
        loading_existing: bool = False,
        
        version = __version__,
        
        *args,
        **kwargs,
    ):
        super().__init__()
        
        name = self._configure_name(name, loading_existing = loading_existing)
        
        self.save_hyperparameters()
        
        # -- torch modules: ----------------------------------------------------
        self._configure_torch_modules(func=PotentialSDE, kwargs=locals())
        self._configure_lightning_model(kwargs = locals())

    def __repr__(self):
        return "LightningSDE-FixedPotential-RegularizedVelocityRatio"




# -- type setting: -------------------------------------------------------------
from typing import Union, List, Optional, Dict


# -- import packages: ----------------------------------------------------------
from tqdm.notebook import tqdm
import lightning
import anndata
import pandas as pd
import torch
import glob
import os
import ABCParse
import pathlib
import autodevice

# -- import local dependencies: ------------------------------------------------
from . import configs, lightning_models, utils, callbacks
from . import _mix_ins as mix_ins
from .. import __version__ # tools

import warnings

warnings.filterwarnings("ignore", ".*Consider increasing the value of the `num_workers` argument*")


# -- model class, faces API: ---------------------------------------------------
class scDiffEq(
    mix_ins.ModelConfigMixIn,
    mix_ins.kNNMixIn,
    mix_ins.StateCharacterizationMixIn,
    mix_ins.UtilityIOMixIn,
    mix_ins.LoggingMixIn,
    mix_ins.PreTrainMixIn,
    mix_ins.TrainMixIn,
    ABCParse.ABCParse,
):
        
    """scDiffeq model class"""
    def __init__(
        self,
        # -- data params: -------------------------------------------------------
        adata: Optional[anndata.AnnData] = None,
        latent_dim: int = 50,
        name: Optional[str] = None,
        use_key: str = "X_pca",
        weight_key: str = 'W',
        obs_keys: List[str] = [],
        seed: int = 0,
        backend: str = "auto",
        gradient_clip_val: float = 0.5,
        # -- velocity ratio keys: -----------------------------------------------
        velocity_ratio_target: float = 0,  # off by default
        velocity_ratio_enforce: float = 0, # off by default
        # -- kNN keys: [optional]: ----------------------------------------------
        build_kNN: Optional[bool] = False,
        kNN_key: Optional[str] = "X_pca",
        kNN_fit_subset: Optional[str] = "train",
        # -- pretrain params: ---------------------------------------------------
        pretrain_epochs: int = 500,
        pretrain_lr: float = 1e-3,
        pretrain_optimizer = torch.optim.Adam,
        pretrain_step_size: int = 100,
        pretrain_scheduler = torch.optim.lr_scheduler.StepLR,
        # -- train params: ------------------------------------------------------
        train_epochs: int = 1500,
        train_lr: float = 1e-5,
        train_optimizer = torch.optim.RMSprop,
        train_scheduler = torch.optim.lr_scheduler.StepLR,
        train_step_size: int = 10,
        train_val_split: List[float] = [0.9, 0.1],
        batch_size: int = 2000,
        train_key: str = "train",
        val_key: str = "val",
        test_key: str = "test",
        predict_key: str = "predict",
        # -- general params: ----------------------------------------------------
        logger: Optional["lightning.logger"] = None,
        num_workers: int = os.cpu_count(),
        silent: bool = True,
        scale_input_counts: bool = False,
        reduce_dimensions: bool = False,
        fate_bias_csv_path: Optional[Union[pathlib.Path, str]] = None,
        fate_bias_multiplier: float = 1,
        viz_frequency: int = 1,
        working_dir: Union[pathlib.Path, str] = os.getcwd(),
        # -- time params: -------------------------------------------------------
        time_key: Optional[str] = None,
        t0_idx: Optional[pd.Index] = None,
        t_min: float = 0,
        t_max: float = 1,
        dt: float = 0.1,
        time_cluster_key: Optional[str] = None,
        t0_cluster: Optional[str] = None,
        shuffle_time_labels: bool = False,
        # -- DiffEq params: ----------------------------------------------------
        mu_hidden: Union[List[int], int] = [400, 400],
        mu_activation: Union[str, List[str]] = "LeakyReLU",
        mu_dropout: Union[float, List[float]] = 0.1,
        mu_bias: bool = True,
        mu_output_bias: bool = True,
        mu_n_augment: int = 0,
        sigma_hidden: Union[List[int], int] = [400, 400],
        sigma_activation: Union[str, List[str]] = "LeakyReLU",
        sigma_dropout: Union[float, List[float]] = 0.1,
        sigma_bias: List[bool] = True,
        sigma_output_bias: bool = True,
        sigma_n_augment: int = 0,
        adjoint: bool = False,
        sde_type: str = "ito",
        noise_type: str = "general",
        brownian_dim: int = 1,
        coef_drift: float = 1.0,
        coef_diffusion: float = 1.0,
        coef_prior_drift: float = 1.0,
        DiffEq_type: str = "SDE",
        potential_type: Union[None, str] = None,
        # other options: "fixed" or "prior"
        # -- Encoder params: ---------------------------------------------------
        encoder_n_hidden: int = 4,
        encoder_power: float = 2,
        encoder_activation: Union[str, List[str]] = "LeakyReLU",
        encoder_dropout: Union[float, List[float]] = 0.2,
        encoder_bias: bool = True,
        encoder_output_bias: bool = True,
        # -- Decoder params: ---------------------------------------------------
        decoder_n_hidden: int = 4,
        decoder_power: float = 2,
        decoder_activation: Union[str, List[str]] = "LeakyReLU",
        decoder_dropout: Union[float, List[float]] = 0.2,
        decoder_bias: bool = True,
        decoder_output_bias: bool = True,
        ckpt_path: Optional[Union[pathlib.Path, str]] = None,
        version: str = __version__,
        *args,
        **kwargs,
    ) -> None:

        """Initialize the scDiffEq model.

        This class is responsible for bringing together three critical
        components: the lightning model, the lightning trainer, and the
        lightning data. All else is superfluous to model operation. 

        Args:
            arg1 (arg1_type): description of arg1. **Default**: None
        
        Returns:
            None
        """
        self.__config__(locals())
        
    def fit(
        self,
        train_epochs=200,
        pretrain_epochs=500,
        train_lr = None,
        pretrain_callbacks: List = [],
        train_callbacks: List = [],
        ckpt_frequency: int = 25,
        save_last_ckpt: bool = True,
        keep_ckpts: int = -1,
        monitor=None,
        accelerator=None,
        log_every_n_steps=1,
        reload_dataloaders_every_n_epochs=1,
        devices=1,
        deterministic=False,
        **kwargs,
    ) -> None:

        """Fit the scDiffEq model to some data.

        Extended description.

        Args:
            arg1 (arg1_type): description of arg1. **Default**: None
        
        Returns:
            None
        """
        
        self.train(**ABCParse.function_kwargs(self.train, locals()))


    def simulate(self) -> anndata.AnnData:
        ...
        
    def __repr__(self) -> str:
        return "scDiffEq"




# -- import packages: ---------------------------------------------------------
import ABCParse
import pathlib
import pandas as pd


# -- import local dependencies: -----------------------------------------------
from ._hparams import HParams
from ._checkpoint import Checkpoint
from ._grouped_metrics import GroupedMetrics


# -- set typing: --------------------------------------------------------------
from typing import Dict, List, Union


# -- operational class: -------------------------------------------------------
class Version(ABCParse.ABCParse):
    """scDiffEq Version object container"""

    def __init__(self, path: Union[pathlib.Path, str] = None, groupby: str = "epoch", *args, **kwargs):
        """Instantiate Version by providing a path

        Args:
            path (Union[pathlib.Path]): path to the version within an
            scDiffEq project. **Default** = None.


        Returns:
            None
        """
        self.__parse__(locals())

    @property
    def _PATH(self) -> pathlib.Path:
        """check and format the provided path"""
        if isinstance(self._path, pathlib.Path):
            return self._path
        if isinstance(self._path, str):
            self._path = pathlib.Path(self._path)
        return self._path

    @property
    def _NAME(self) -> str:
        """version name from provided path"""
        return self._PATH.name

    @property
    def _CONTENTS(self) -> List[pathlib.Path]:
        """return one-level glob of the provided path"""
        return list(self._PATH.glob("*"))

    @property
    def hparams(self):
        """check if the .yaml exists and instantiate the HParams cls each time"""
        hparams_path = self._PATH.joinpath("hparams.yaml")
        if hparams_path.exists():
            return HParams(hparams_path)

    @property
    def metrics_df(self) -> pd.DataFrame:
        """check if metrics.csv path exists as given and read. reads new
        each time"""
        metrics_path = self._PATH.joinpath("metrics.csv")
        if metrics_path.exists():
            return pd.read_csv(metrics_path)
        
    @property
    def per_epoch_metrics(self):
        self._GROUPED_METRICS = GroupedMetrics(groupby = self._groupby)
        return self._GROUPED_METRICS(self.metrics_df)

    @property
    def _CKPT_PATHS(self) -> List[pathlib.Path]:
        """formatted ckpt paths"""
        _ckpt_paths = list(self._PATH.joinpath("checkpoints").glob("*"))
        return [pathlib.Path(path) for path in _ckpt_paths]

    @property
    def _SORTED_CKPT_KEYS(self) -> List:
        """sorting for organization's sake"""
        epochs = list(self.ckpts.keys())
        _epochs = sorted([epoch for epoch in epochs if epoch != "last"])
        if "last" in epochs:
            _epochs.append("last")
        return _epochs

    @property
    def ckpts(self) -> Dict:
        """format and update available checkpoints for the version"""
        if not hasattr(self, "_CHECKPOINTS"):
            self._CHECKPOINTS = {}
            for ckpt_path in self._CKPT_PATHS:
                ckpt = Checkpoint(ckpt_path)
                self._CHECKPOINTS[ckpt.epoch] = ckpt
        return self._CHECKPOINTS

    def __repr__(self) -> str:
        """return the name of the obj"""
        return self._NAME




__module_name__ = "__init__.py"
__doc__ = """tools __init__ module. Sub-package of the main scdiffeq API."""
__author__ = ", ".join(["Michael E. Vinyard",])
__email__ = ", ".join(["mvinyard.ai@gmail.org",])


from . import utils

from ._grid_velocity import GridVelocity
from ._velocity_embedding import VelocityEmbedding
from ._velocity_graph import velocity_graph

# import functions accessed as sdq.tl.<func>: --------------------------------------------
from ._annotate_cells import annotate_cells
from ._time_free_sampling import time_free_sampling

from ._reconstruct_function import reconstruct_function
from ._versions import Versions, configure_version
# from ._func_from_version import func_from_version
from ._umap import UMAP

from ._drift_diffusion_state_characterization import drift, diffusion

from ._dimension_reduction import DimensionReduction


# -----------

# from ._x_use import X_use, fetch_formatted_data
from ._knn import kNN
from ._knn_smoothing import kNNSmoothing

from ._negative_cross_entropy import NegativeCrossEntropy

from ._sum_norm_df import sum_norm_df


from ._feature_correlation import (
    FeatureCorrelation,
    drift_correlated_features,
    diffusion_correlated_features,
    potential_correlated_features,
)

from ._cell_potential import cell_potential, normalize_cell_potential

from ._final_state_per_simulation import FinalStatePerSimulation

# from ._simulator import Simulator, simulate
from ._simulation import Simulation, simulate

from ._perturbation import Perturbation, perturb

from ._grouped_expression import GroupedExpression, grouped_expression
from ._annotate_gene_features import GeneCompatibility, annotate_gene_features
# from ._compared_temporal_gene_expression import TemporalGeneExpression, compared_temporal_gene_expression
# from ._smoothed_expression import SmoothedExpression, smoothed_expression
from ._temporal_expression import temporal_expression

from ._annotate_cell_state import annotate_cell_state
from ._annotate_cell_fate import annotate_cell_fate
from ._invert_scaled_gex import invert_scaled_gex
from ._fate_perturbation_experiment import FatePerturbationExperiment
from ._fate_perturbation_screen import FatePerturbationScreen




import lightning
import anndata

from .. import configs, utils

from typing import Dict, Optional

class ModelConfigMixIn(object):
    """configure_model and configure_data can/should be accessed from the public handle"""

    def _configure_trainer_generator(self):
        
        """ """
                
        self.TrainerGenerator = configs.LightningTrainerConfiguration(
            save_dir=self._name
        )

    def configure_model(
        self,
        DiffEq: Optional[lightning.LightningModule] = None,
        configure_trainer: bool = True,
        loading_existing: bool = False,
    ):
        if DiffEq is None:
            self._LitModelConfig = configs.LightningModelConfiguration(
                data_dim=self._data_dim,
                latent_dim=self._latent_dim,
                DiffEq_type=self._DiffEq_type,
                potential_type=self._potential_type,
                fate_bias_csv_path=self._fate_bias_csv_path,
                velocity_ratio_target = self._velocity_ratio_target,
            )

            if hasattr(self, "reducer"):
                self._PARAMS["PCA"] = self.reducer.PCA
            DiffEq = self._LitModelConfig(self._PARAMS, self._ckpt_path, loading_existing = loading_existing)

        self.DiffEq = DiffEq
        self._name = self.DiffEq.hparams.name

        self._INFO(f"Using the specified parameters, {self.DiffEq} has been called.")
        self._component_loader = utils.FlexibleComponentLoader(self)

        lightning.seed_everything(self._seed)

        # -- Step 5: configure bridge to lightning logger ----------------------
        # was its own step before: now in-line here, since it
        # doesn't make sense to separate it, functionally
        self._LOGGING = utils.LoggerBridge(self.DiffEq)

        if configure_trainer:
            self._configure_trainer_generator()

    def configure_data(self, adata: anndata.AnnData):
        """ """
        self.adata = adata.copy()
        
        self._DATA_CONFIG = configs.DataConfiguration()
        self._DATA_CONFIG(scDiffEq = self)
        self._INFO(f"Input data configured.")
        
    def __config__(self, kwargs: Dict):
        """ """
        # -- Step 1: parse kwargs, set up info msg -----------------------------
        self.__parse__(kwargs, public = [None], ignore=["adata"])

        # -- Step 2: configure data ----------------------------------------------
        if not kwargs['adata'] is None:
            # if adata is given, triggers 2 through 4
            self.configure_data(adata = kwargs['adata'])

            # -- Step 3: configure kNN --------------------------------------------
            if self._PARAMS["build_kNN"]:
                self._PARAMS['kNN'] = self.kNN

            # -- Step 4: configure model -------------------------------------------
            self.configure_model(DiffEq = None, configure_trainer = True)

            # -- Step 5: extras (was step 6): ---------------------------------------
#             if kwargs["reduce_dimensions"]:
#                 self._configure_dimension_reduction()




import anndata
import ABCParse
import pandas as pd
import cell_perturb
import scipy.stats
import numpy as np
import lightning

from ._simulation import simulate

from ._annotate_cell_state import annotate_cell_state
from ._annotate_cell_fate import annotate_cell_fate

from typing import Optional, List, Dict, Union


class PerturbationExperimentResult(ABCParse.ABCParse):
    def __init__(
        self,
        ctrl_result: anndata.AnnData,
        prtb_result: anndata.AnnData,
        *args,
        **kwargs,
    ):
        """
        Args:
            ctrl_result (anndata.AnnData)

            prtb_result (anndata.AnnData)
        """

        self.__parse__(locals())

    def _count_fates(self, result: pd.DataFrame):

        fate_counts = (
            result.obs.groupby("replicate")["fate"].value_counts().unstack().fillna(0)
        )
        return fate_counts.div(fate_counts.sum(1), axis=0).T

    def _forward(self, result: anndata.AnnData, key: str):
        return self._count_fates(result)

    #         result.columns = [f"{key}.{i}" for i in result.columns.tolist()]

    @property
    def ctrl(self) -> pd.DataFrame:
        if not hasattr(self, "_ctrl"):
            self._ctrl = self._forward(self._ctrl_result, key="ctrl")
        return self._ctrl

    @property
    def prtb(self) -> pd.DataFrame:
        if not hasattr(self, "_prtb"):
            self._prtb = self._forward(self._prtb_result, key="prtb")
        return self._prtb

    @property
    def _fates(self):
        return list(set(self.ctrl.index.tolist()).union(self.prtb.index.tolist()))

    def _zerofill(self, result_df: pd.DataFrame):
        result_t = result_df.copy().T
        for fate in self._fates:
            if not fate in result_t:
                result_t[fate] = 0

        return result_t

    def _compute_lfc(self, ctrl_t, prtb_t, constant: float = 1e-9):
        return (prtb_t + constant).div((ctrl_t + constant)).apply(np.log2)

    def _compute_pvals(self, ctrl_t, prtb_t):
        return pd.Series(
            {
                fate: scipy.stats.ttest_ind(
                    ctrl_t[fate], prtb_t[fate], equal_var=False
                )[1]
                for fate in self._fates
            }
        )

    def _compute_summary_statistics(self):
        
        ctrl_t = self._zerofill(self.ctrl)
        prtb_t = self._zerofill(self.prtb)
        self._lfc = self._compute_lfc(ctrl_t, prtb_t)
        pvals = self._compute_pvals(ctrl_t, prtb_t)
        
        lfc_pvals = pd.DataFrame([self._lfc.mean(), self._lfc.std(), pvals]).T
        lfc_pvals.columns = ["lfc", 'lfc_std', "pval"]

        return lfc_pvals
    
    @property
    def stats(self):
        if not hasattr(self, "_stats"):
            self._stats = self._compute_summary_statistics()
        return self._stats

    def __repr__(self):
        return "PerturbationExperimentResult"


class FatePerturbationExperiment(ABCParse.ABCParse):
    """Container class for an expression perturbation experiment"""
    def __init__(
        self,
        seed: int = 0,
        use_key: str = "X_scaled",
        replicates: int = 5,
        N: int = 200,
        time_key: str = "t",
        *args,
        **kwargs,
    ):
        self.__parse__(locals())

    @property
    def _PERTURBATION_INIT_KWARGS(self) -> Dict:
        """function kwargs for ``cell_perturb.Perturbation.__init__``."""
        return ABCParse.function_kwargs(
            func=cell_perturb.Perturbation.__init__, kwargs=self._PARAMS
        )

    @property
    def _PERTURBATION_CALL_KWARGS(self):
        """function kwargs for ``cell_perturb.Perturbation.__call__``."""
        return ABCParse.function_kwargs(
            func=cell_perturb.Perturbation.__call__, kwargs=self._PARAMS
        )

    @property
    def adata_prtb(self) -> anndata.AnnData:
        """adata_prtb (anndata.AnnData)"""
        if not hasattr(self, "_adata_prtb"):
            self._perturbation = cell_perturb.Perturbation(
                **self._PERTURBATION_INIT_KWARGS
            )
            self._adata_prtb = self._perturbation(**self._PERTURBATION_CALL_KWARGS)
        return self._adata_prtb

    def _subset_final_state(self, adata_sim) -> anndata.AnnData:
        """adata at final state"""
        t = adata_sim.obs[self._time_key]
        return adata_sim[t == t.max()].copy()
    
    @property
    def DiffEq(self) -> lightning.LightningModule:
        """DiffEq (lightning.LightningModule)"""
        if isinstance(self._model.DiffEq, lightning.LightningModule):
            return self._model.DiffEq
        elif isinstance(self._model, lightning.LightningModule):
            return self._model    

    def forward(self):
        """run simulation ctrl vs. prtb"""
        adata_sim_prtb = simulate(
            adata=self.adata_prtb,
            diffeq=self.DiffEq,
            use_key="X_pca_prtb",
            time_key = self._time_key,
        )
        adata_sim_ctrl = simulate(
            adata=self.adata_prtb,
            diffeq=self.DiffEq,
            use_key="X_pca_ctrl",
            time_key = self._time_key,
        )
        prtb = self._subset_final_state(adata_sim_prtb)
        ctrl = self._subset_final_state(adata_sim_ctrl)
        
        rep = self.adata_prtb.obs["replicate"].values
        
        prtb.obs["replicate"] = rep
        ctrl.obs["replicate"] = rep

        annotate_cell_state(prtb, self._model.kNN, silent=True)
        annotate_cell_fate(prtb, silent=True)

        annotate_cell_state(ctrl, self._model.kNN, silent=True)
        annotate_cell_fate(ctrl, silent=True)
        
        return ctrl, prtb

    def __call__(
        self,
        adata: anndata.AnnData,
        model: "scdiffeq.scDiffEq",
        genes: List[str],
        subset_key: str,
        subset_val: str,
        target_value: float = 10,
        PCA: Optional = None,
        *args,
        **kwargs,
    ):
        """
        Run perturbation screen.
        
        Args:
            adata (anndata.AnnData): adata obj.
            
            
            model ("scdiffeq.scDiffEq"): scDiffEq model.
            
            genes (List[str]): Genes over which screen should be run.
        
            subset_key (str):
            
            subset_val (str):
        
            target_value (float): Z-score value at which perturbation should be set. **Default**: 10
            
            PCA (Optional[sklearn.decomposition.PCA]: PCA model for transforming expression to model input. **Default**: None.
        
        Returns:
            PerturbationExperimentResult
        """
        self.__update__(locals())

        self.ctrl_result, self.prtb_result = self.forward()
        
        return PerturbationExperimentResult(self.ctrl_result, self.prtb_result)




__module_name__ = "_checkpoint.py"
__doc__ = """Module containing the ``Checkpoint`` container object."""
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["mvinyard.ai@gmail.com"])


# -- import packages: ---------------------------------------------------------
import ABCParse
import pathlib
import torch


# -- set typing: --------------------------------------------------------------
from typing import Union, Dict


# -- operational class: -------------------------------------------------------
class Checkpoint(ABCParse.ABCParse):
    def __init__(self, path: Union[pathlib.Path, str], *args, **kwargs) -> None:
        """Instantiates checkpoint
        
        Args:
            path (Union[pathlib.Path, str]): path to saved checkpoint.
            
        Returns:
            None
        """
        self.__parse__(locals())
        
    @property
    def path(self) -> pathlib.Path:
        """"""
        return pathlib.Path(self._path)

    @property
    def _fname(self) -> str:
        """filename"""
        return self.path.name.split(".")[0]

    @property
    def epoch(self) -> Union[int, str]:
        if self._fname != "last":
            return int(self._fname.split("=")[1].split("-")[0])
        return self._fname

    @property
    def state_dict(self) -> Dict[str, 'LightningCheckpoint']:
        """loads state_dict created by PyTorch Lightning"""
        if not hasattr(self, "_ckpt"):
            self._state_dict = torch.load(self.path)  # ["state_dict"]
        return self._state_dict

    def __repr__(self) -> str:
        """obj description of checkpoint at epoch"""
        return f"ckpt epoch: {self.epoch}"




import torch

from ....tools.utils import L2Norm

class RegularizedVelocityRatioMixIn(object):
    def __init__(self):
        """ """
        super().__init__()
        
        self._L2Norm = L2Norm()


    def forward(self, Z0, t, **kwargs):
        """Forward step: (0) integrate in latent space"""
        return self.integrate(Z0=Z0, t=t, dt=self.hparams["dt"], logqp=False, **kwargs)
    
    
    def _compute_magnitude_f(self, Z_hat: torch.Tensor) -> torch.Tensor:
        return self._L2Norm(self.DiffEq.drift(Z_hat)).mean(1) * self.hparams['dt']
        
    def _compute_magnitude_g(self, Z_hat: torch.Tensor) -> torch.Tensor:
        return self._L2Norm(torch.stack([self.DiffEq.diffusion(Z) for Z in Z_hat]).squeeze(-1)).mean(1)
    
    def _velocity_ratio_loss(self, Z_hat: torch.Tensor) -> torch.Tensor:
        """ """
        
        TARGET_RATIO = self.hparams['velocity_ratio_target']
        VELO_RATIO_ENFORCE = self.hparams['velocity_ratio_enforce']
        
        self.reg_f = self._compute_magnitude_f(Z_hat)
        self.reg_g = self._compute_magnitude_g(Z_hat)
        
        VELO_RATIO_LOSS = VELO_RATIO_ENFORCE * (TARGET_RATIO - self.reg_f.div(self.reg_g))
        return torch.abs(VELO_RATIO_LOSS)
    
    def step(self, batch, batch_idx, stage=None):

        batch = self.process_batch(batch, batch_idx)
        X_hat = self.forward(batch.X0, batch.t)
        self.sinkhorn_loss = self.compute_sinkhorn_divergence(
            X = batch.X, X_hat = X_hat, W = batch.W, W_hat = batch.W_hat
        )
        
        self.velocity_ratio_loss = self._velocity_ratio_loss(X_hat)
        
        if self.hparams['disable_velocity_ratio_backprop']:
            self.total_loss = self.sinkhorn_loss
        else:
            self.total_loss = self.sinkhorn_loss + self.velocity_ratio_loss

        return self.total_loss.sum()




class BaseForwardMixIn(object):
    def __init__(self):
        super().__init__()


    def forward(self, Z0, t, **kwargs):
        """Forward step: (0) integrate in latent space"""
        return self.integrate(Z0=Z0, t=t, dt=self.hparams["dt"], logqp=False, **kwargs)
    
    def step(self, batch, batch_idx, stage=None):

        batch = self.process_batch(batch, batch_idx)
        X_hat = self.forward(batch.X0, batch.t)
        self.sinkhorn_loss = self.compute_sinkhorn_divergence(
            X = batch.X, X_hat = X_hat, W = batch.W, W_hat = batch.W_hat
        )
        return self.sinkhorn_loss.sum()




import anndata
import ABCParse
from typing import Optional

from ... import tools

class kNNMixIn(ABCParse.ABCParse, object):
    def __init__(self, *args, **kwargs):
        self.__parse__(locals())
        
    @property
    def _adata_kNN_fit(self):
        return self.adata[self.adata.obs[self._kNN_fit_subset]].copy()
        
    def _kNN_info_msg(self):
        
        self._INFO(f"Bulding Annoy kNN Graph on adata.obsm['{self._kNN_fit_subset}']")
        
    def configure_kNN(
        self,
        adata: Optional[anndata.AnnData] = None,
        kNN_key: Optional[str] = None,
        kNN_fit_subset: Optional[str] = None,
    ):
        
        """
        subset key should point to a col in adata.obs of bool vals
        """
        self.__update__(locals())
        
        self._kNN_info_msg()
        self._kNN = tools.kNN(
            adata = self._adata_kNN_fit, use_key = self._kNN_key,
        )

    @property
    def kNN(self):
        if not hasattr(self, "_kNN"):
            self.configure_kNN(
                adata = self._adata_kNN_fit,
                kNN_key = self._kNN_key,
                kNN_fit_subset = self._kNN_fit_subset,
            )
        return self._kNN





import ABCParse
import anndata
import pathlib
import lightning
import torch
import torch_nets
import yaml
import pandas as pd

from ...core import lightning_models, utils, scDiffEq


from typing import Optional, Union

from ._hparams import HParams
from ._project import Project
from ._checkpoint import Checkpoint
from ._version import Version


class ModelLoader(ABCParse.ABCParse):
    def __init__(self, project_path = None, version = None, name_delim: str = ".", *args, **kwargs):
        """ """
        self.__parse__(locals())
        self._validate_version()

    @property
    def project(self):
        if not hasattr(self, "_project"):
            self._project = Project(path=self._project_path)
        return self._project

    @property
    def _VERSION_KEY(self):

        if not hasattr(self, "_VERSION_KEY_PRIVATE"):
            if self._version is None:
                version_key = max(self.project._VERSION_PATHS)
                self._INFO(f"Version not provided. Defaulting to: '{version_key}'")
                self._VERSION_KEY_PRIVATE = version_key
            else:
                self._VERSION_KEY_PRIVATE = f"version_{self._version}"

        return self._VERSION_KEY_PRIVATE

    @property
    def _VERSION_PATH(self):
        return self.project._VERSION_PATHS[self._VERSION_KEY]

    @property
    def version(self):
        if not hasattr(self, "_VERSION"):
            self._VERSION = Version(path=self._VERSION_PATH)
        return self._VERSION

    @property
    def hparams(self):
        return self.version.hparams()

    @property
    def _MODEL_TYPE(self):
        name = self.hparams["name"]
        if ":" in name: # for backwards-compatibility
            return name.split(":")[0].replace("-", "_")
        return name.split(self._name_delim)[0].replace("-", "_")

    @property
    def LightningModule(self):
        return getattr(lightning_models, self._MODEL_TYPE)
#         if not hasattr(self, "_model"):
#             LitModel = getattr(lightning_models, self._MODEL_TYPE)
#             self._model = LitModel(**self.hparams)
#         return self._model

    @property
    def ckpt(self):
        return self.version.ckpts[self.epoch]
    
    @property
    def epoch(self):
        if not hasattr(self, "_epoch"):
            self._epoch = "last"
            self._INFO(f"Epoch not provided. Defaulting to: '{self._epoch}'")
        return self._epoch

    def _validate_epoch(self):
            
        msg = f"ckpt at epoch: {self.epoch} does not exist. Choose from: {self.version._SORTED_CKPT_KEYS}"
        assert self.epoch in self.version.ckpts, msg

    def _validate_version(self):
        
        version_key = self._VERSION_KEY
        available_versions = [
            attr for attr in self.project.__dir__() if attr.startswith("version_")
        ]

        msg = f"{version_key} not found in project. Available versions: {available_versions}"
        assert version_key in self.project._VERSION_PATHS, msg

    def from_ckpt(self, ckpt_path: str, load_kwargs = {"loading_existing": True}):
        return self.LightningModule.load_from_checkpoint(ckpt_path, **load_kwargs)
    
    def __call__(
        self,
        epoch: Optional[Union[str, int]] = None,
        plot_state_change: Optional[bool] = False,
        load_kwargs = {"loading_existing": True},
        *args,
        **kwargs,
    ) -> lightning.LightningModule:
        
        """
        Parameters
        ----------
        epoch: Optional[Union[str, int]], default = None
        
        plot_state_change: Optional[bool], default = False
        
        Returns
        -------
        model: 
        """
        
        self.__update__(locals())

        self._validate_epoch()
        ckpt_path = self.ckpt.path
        return self.LightningModule.load_from_checkpoint(ckpt_path, **load_kwargs)

#         if plot_state_change:
#             torch_nets.pl.weights_and_biases(model.state_dict())
            
#         ckpt_path = self.ckpt.path
#         self._INFO(f"Loading model from ckpt: \n\t'{ckpt_path}'")
#         model = model.load_from_checkpoint(ckpt_path)
        
#         if plot_state_change:
#             torch_nets.pl.weights_and_biases(model.state_dict())

def _inputs_from_ckpt_path(ckpt_path):
    """If you give the whole ckpt_path, you can derive the other inputs."""
    
    if isinstance(ckpt_path, str):
        ckpt_path = pathlib.Path(ckpt_path)

    project_path = ckpt_path.parent.parent.parent
    version = int(ckpt_path.parent.parent.name.split("_")[1])
#     epoch = int(ckpt_path.name.split("-")[0].split("=")[1])
    
    return {"project_path": project_path, "version": version} # , "epoch": epoch}
    
def load_diffeq(
    ckpt_path: Optional[Union[pathlib.Path, str]] = None,
    project_path: Optional[Union[pathlib.Path, str]] = None,
    version: Optional[int] = None,
    epoch: Optional[Union[int, str]] = None,
) -> lightning.LightningModule:

    """
    Load DiffEq from project_path, version [optional], and epoch [optional].
    
    Args:
        project_path (Union[pathlib.Path, str])

        version (Optional[int]): **Default** = None

        epoch (Optional[Union[int, str]]): **Default** = None

    
    Returns:
        DiffEq (lightning.LightningModule): lightning differential equation model.
    """
    
    if not ckpt_path is None:
        inputs = _inputs_from_ckpt_path(ckpt_path)
        project_path = inputs['project_path']
        version = inputs['version']
#         epoch = inputs['epoch']

    model_loader = ModelLoader(project_path=project_path, version=version)
    
    if not ckpt_path is None:
        return model_loader.from_ckpt(ckpt_path)
    return model_loader(epoch=epoch)

def load_model(
    adata: anndata.AnnData,
    ckpt_path: Optional[Union[pathlib.Path, str]] = None,
    project_path: Optional[Union[pathlib.Path, str]] = None,
    version: Optional[int] = None,
    epoch: Optional[Union[int, str]] = None,
    configure_trainer: Optional[bool] = False,
):
    
    """Load scDiffEq model.
    
    Args:
        adata (anndata.AnnData): adata object.
        
        ckpt_path (Optional[Union[pathlib.Path, str]]): description. **Default** = None
        
        project_path (Optional[Union[pathlib.Path, str]]): description. **Default** = None
        
        version (Optional[int]): description. **Default** = None
        
        epoch (Optional[Union[int, str]]): description. **Default** = None
        
        configure_trainer (Optional[bool]): indicate if trainer should be configured. **Default** = False.
    
    Returns:
        scdiffeq.scDiffEq
    """

    diffeq = load_diffeq(
        ckpt_path=ckpt_path,
        project_path=project_path,
        version=version,
        epoch=epoch,
    )
    
    model = scDiffEq(**dict(diffeq.hparams))
    model._load_version = version
    model.configure_data(adata)
    model.configure_kNN()
    model.configure_model(
        diffeq, configure_trainer = configure_trainer, loading_existing = True,
    )
    
    PREVIOUS_METRICS_PATH = pathlib.Path(project_path).joinpath(f"version_{version}/metrics.csv")
    PREVIOUS_METRICS = pd.read_csv(PREVIOUS_METRICS_PATH)
    
    model.DiffEq.COMPLETED_EPOCHS = int(PREVIOUS_METRICS["total_epochs"].max())

    return model




# -- import packages: ---------------------------------------------------------
import ABCParse
import pathlib


# -- import local dependencies: -----------------------------------------------
from ._version import Version


# -- set typing: --------------------------------------------------------------
from typing import Dict, Union


# -- operational class: -------------------------------------------------------
class Project(ABCParse.ABCParse):
    """Object container for an scDiffEq project"""

    def __init__(
        self,
        path: Union[str, pathlib.Path] = pathlib.Path("./").absolute(),
        metrics_groupby: str = "epoch",
        *args,
        **kwargs,
    ):
        """Initialize the project object by providing a path.

        Args:
            path (pathlib.Path): path to the project, created by scDiffeq.

        Returns:
            None
        """
        self.__parse__(locals())

        self._set_version_path_attributes()

    @property
    def _PROJECT_PATH(self) -> pathlib.Path:
        """check and format the provided project path"""
        if isinstance(self._path, pathlib.Path):
            return self._path
        elif isinstance(self._path, str):
            return pathlib.Path(self._path)
        else:
            raise TypeError("arg: `path` must be of type: [pathlib.Path, str]")

    @property
    def _VERSION_PATHS(self) -> Dict[str, pathlib.Path]:
        """Assemble available version paths based on the provided project path"""
        version_paths = sorted(list(self._PROJECT_PATH.glob("version_*")))
        return {path.name: path for path in version_paths}

    def _set_version_path_attributes(self) -> None:
        """sets the name of each version as a cls attribute, pointing to the
        path of the version."""
        for v_name, v_path in self._VERSION_PATHS.items():
            version = Version(path = v_path, groupby = self._metrics_groupby)
            setattr(self, v_name, version)
            
    def __getitem__(self, version: int) -> pathlib.Path:
        """format version key and return version"""
        return getattr(self, f"version_{version}")

    def __repr__(self) -> str:
        """return the name of the obj"""
        return """scDiffEq Project"""



# -- import packages: ---------------------------------------------------------
import ABCParse
import anndata
import matplotlib.pyplot as plt
import numpy as np
import cellplots as cp

import matplotlib.cm

# -- import local dependencies: -----------------------------------------------
from ..tools import VelocityEmbedding, GridVelocity


# -- type setting: ------------------------------------------------------------
from typing import Any, Dict, List, Optional, Union, Tuple


# -- Operational class: -------------------------------------------------------
class VelocityStreamPlot(ABCParse.ABCParse):
    """Velocity stream plot for a single ax"""

    def __init__(
        self,
        density: float = 1,
        smooth: float = 0.5,
        n_neighbors: Optional[int] = None,
        min_mass: float = 1,
        autoscale: bool = True,
        stream_adjust: bool = True,
        cutoff_percentile: float = 0.05,
        velocity_key: str = "velocity",
        self_transitions: bool = True,
        use_negative_cosines: bool = True,
        T_scale: float = 10,
        *args,
        **kwargs,
    ):
        self.__parse__(locals())

        self._velocity_emb = VelocityEmbedding(
            velocity_key=velocity_key,
            self_transitions=self_transitions,
            use_negative_cosines=use_negative_cosines,
            T_scale=T_scale,
        )
        self._grid_velocity = GridVelocity(
            density=density,
            smooth=smooth,
            n_neighbors=n_neighbors,
            min_mass=min_mass,
            autoscale=autoscale,
            stream_adjust=stream_adjust,
            cutoff_percentile=cutoff_percentile,
        )


    @property
    def X_emb(self):
        if not hasattr(self, "_X_emb"):
            self._X_emb, self._V_emb = self._velocity_emb(self._adata)
        return self._X_emb

    @property
    def V_emb(self):
        if not hasattr(self, "_V_emb"):
            self._X_emb, self._V_emb = self._velocity_emb(self._adata)
        return self._V_emb

    @property
    def X_grid(self):
        if not hasattr(self, "_X_grid"):
            self._X_grid, self._V_grid = self._grid_velocity(self.X_emb, self.V_emb)
        return self._X_grid

    @property
    def V_grid(self):
        if not hasattr(self, "_V_grid"):
            self._X_grid, self._V_grid = self._grid_velocity(self.X_emb, self.V_emb)
        return self._V_grid

    @property
    def x(self):
        return self.X_grid[0]

    @property
    def y(self):
        return self.X_grid[1]

    @property
    def u(self):
        return self.V_grid[0]

    @property
    def v(self):
        return self.V_grid[1]

    @property
    def xmin(self):
        return np.min(self.X_emb[:, 0])

    @property
    def xmax(self):
        return np.max(self.X_emb[:, 0])

    @property
    def ymin(self):
        return np.min(self.X_emb[:, 1])

    @property
    def ymax(self):
        return np.max(self.X_emb[:, 1])

    @property
    def xmargin(self):
        return (self.xmax - self.xmin) * self._add_margin

    @property
    def ymargin(self):
        return (self.ymax - self.ymin) * self._add_margin

    def _set_margin(self, ax):
        """"""
        ax.set_xlim(self.xmin - self.xmargin, self.xmax + self.xmargin)
        ax.set_ylim(self.ymin - self.ymargin, self.ymax + self.ymargin)

    @property
    def _STREAMPLOT_KWARGS(self) -> Dict[str, Any]:
        kwargs = {
            "color": self._stream_color,
            "density": self._stream_density,
            "linewidth": self._linewidth,
            "zorder": self._stream_zorder,
            "arrowsize": self._arrowsize,
            "arrowstyle": self._arrowstyle,
            "maxlength": self._maxlength,
            "integration_direction": self._integration_direction,
        }
        kwargs.update(self._stream_kwargs)
        return kwargs

    def streamplot(self, ax) -> None:
        ax.streamplot(self.x, self.y, self.u, self.v, **self._STREAMPLOT_KWARGS)
        self._set_margin(ax)
        
    @property
    def _SCATTER_KWARGS(self) -> Dict[str, Any]:
        """ """
        kwargs = {
            "c": self._c,
            "zorder": self._scatter_zorder,
            "ec": "None",
            "alpha": 0.2,
            "s": 50,
        }
        kwargs.update(self._scatter_kwargs)
        return kwargs
    
    def _SCATTER_CMAP(self, groups) -> Dict:
        """ """
        if not hasattr(self, "_cmap"):
            print("no cmap found...")
            self._cmap = matplotlib.cm.tab20.colors
        if not isinstance(self._cmap, Dict):
            print("cmap found...not a dictionary")
            self._cmap = {group: self._cmap[en] for en, group in enumerate(groups)}
        return self._cmap

    def scatter(self, ax) -> None:
        """ """
        obs_df = self._adata.obs.copy().reset_index()
        cols = obs_df.columns.tolist()

        kwargs = self._SCATTER_KWARGS

        COLOR_FROM_OBS = self._c in cols
        COLOR_BY_GROUP = str(obs_df[self._c].dtype) == "categorical"
        
        if COLOR_FROM_OBS:
            if not COLOR_BY_GROUP: # implies float not grouped object.
                kwargs.update({"c":obs_df[self._c]})
                ax.scatter(self.X_emb[:, 0], self.X_emb[:, 1], **kwargs)
            else:
                kwargs.pop("c")
                groups = obs_df.groupby(self._c).groups # dict
                cmap = self._SCATTER_CMAP(groups)
                for group, group_ix in groups.items():
                    if hasattr(self, "_group_zorder") and group in self._group_zorder:
                        kwargs.update({'zorder': self._group_zorder[group]})
                    ax.scatter(
                        self.X_emb[group_ix, 0], self.X_emb[group_ix, 1], color = cmap[group], **kwargs,
                    )

    def __call__(
        self,
        adata: anndata.AnnData,
        ax: Optional[Union[plt.Axes, List[plt.Axes]]] = None,
        stream_color: str = "k",
        c: str = "dodgerblue",
        group_zorder: Optional[Dict] = None,
        cmap: Optional[Union[Dict,List,Tuple]] = 'matplotlib.cm.tab20.colors',
        linewidth: float = 0.5,
        stream_density: float = 2.5,
        add_margin: float = 0.1,
        arrowsize: float = 1,
        density: float = 1,
        arrowstyle: str = "-|>",
        maxlength: float = 4,
        integration_direction: str = "both",
        scatter_zorder: int = 0,
        stream_zorder: int = 10,
        mpl_kwargs: Optional[Dict] = {},
        scatter_kwargs: Optional[Dict] = {},
        stream_kwargs: Optional[Dict] = {},
        disable_scatter: bool = False,
        *args,
        **kwargs,
    ):
        """
        Args:

        Returns
            None
        """
        self.__update__(locals())
        
        if ax is None:
            _mpl_kwargs = {
                "nplots": 1,
                "ncols": 1,
                "height": 1.,
                "width": 1.,
                "delete": "all",
                "del_xy_ticks": [True],
            }
            _mpl_kwargs.update(mpl_kwargs)
            fig, axes = cp.plot(**_mpl_kwargs)

        else:
            axes = ABCParse.as_list(ax)
        
        for ax in axes:
            self.streamplot(ax)
            if not self._disable_scatter:
                self.scatter(ax)

# -- API-facing function: -----------------------------------------------------
def velocity_stream(
    adata: anndata.AnnData,
    ax: Optional[Union[plt.Axes, List[plt.Axes]]] = None,
    c: str = "dodgerblue",
    cmap: Optional[Union[Dict,List,Tuple]] = matplotlib.cm.tab20.colors,
    group_zorder: Optional[Dict] = None,
    linewidth: float = 0.5,
    stream_density: float = 2.5,
    add_margin: float = 0.1,
    arrowsize: float = 1,
    arrowstyle: str = "-|>",
    maxlength: float = 4,
    integration_direction: str = "both",
    scatter_zorder: int = 0,
    stream_zorder: int = 10,
    density: float = 1,
    smooth: float = 0.5,
    n_neighbors: Optional[int] = None,
    min_mass: float = 1,
    autoscale=True,
    stream_adjust=True,
    cutoff_percentile: float = 0.05,
    velocity_key: str = "velocity",
    self_transitions: bool = True,
    use_negative_cosines: bool = True,
    T_scale: float = 10,
    stream_kwargs: Optional[Dict[str, Any]] = {},
    scatter_kwargs: Optional[Dict[str, Any]] = {},
    mpl_kwargs: Optional[Dict[str, Any]] = {},
    *args,
    **kwargs,
):
    """"""

    init_kwargs = ABCParse.function_kwargs(VelocityStreamPlot.__init__, locals())
    call_kwargs = ABCParse.function_kwargs(VelocityStreamPlot.__call__, locals())
    velo_stream_plot = VelocityStreamPlot(**init_kwargs)
    velo_stream_plot(**call_kwargs)




# -- import packages: ---------------------------------------------------------
import anndata
import ABCParse
import numpy as np
import pandas as pd


# -- import local dependencies: -----------------------------------------------


# -- set typing: --------------------------------------------------------------



# -- controlling class: -------------------------------------------------------
class CellFateAnnotation(ABCParse.ABCParse):
    """ """
    def __init__(self, time_key: str = "t", sim_key: str = "sim", silent: bool = False, *args, **kwargs):
        self.__parse__(locals())

    @property
    def _TIME(self) -> pd.Series:
        return self._adata_sim.obs[self._time_key]

    @property
    def _T_MAX(self) -> float:
        """Max time observed."""
        return self._TIME.max()

    @property
    def _FATE_VALUES(self) -> np.ndarray:
        return self._adata_sim.obs.loc[self._TIME == self._T_MAX][
            [self._sim_key, self._state_key]
        ].values

    def forward(self) -> None:
        mappable_fate_dict = {row[0]: row[1] for row in self._FATE_VALUES}
        self._adata_sim.obs[self._key_added] = self._adata_sim.obs[self._sim_key].map(
            mappable_fate_dict
        )
        if not self._silent:
            self._INFO(f"Added fate annotation: adata_sim.obs['{self._key_added}']")
        self._count_fates()

    def _count_fates(self) -> None:
        """ """
        fate_obs = self._adata_sim.obs.loc[self._TIME == self._T_MAX]
        self._adata_sim.uns['fate_counts'] = fate_obs[self._key_added].value_counts().to_dict()
        
        if not self._silent:
            self._INFO("Added fate counts: adata_sim.uns['fate_counts']")
        
    def __call__(
        self,
        adata_sim: anndata.AnnData,
        state_key: str,
        key_added: str = "fate",
        *args,
        **kwargs,
    ) -> None:
        self.__update__(locals())

        self.forward()


# -- API-facing function: -----------------------------------------------------
def annotate_cell_fate(
    adata_sim: anndata.AnnData,
    state_key: str = "state",
    key_added: str = "fate",
    time_key: str = "t",
    sim_key: str = "sim",
    silent: bool = False, 
    *args,
    **kwargs,
) -> None:

    """Annotate the fate of a simulated cell trajectory, given annotated states.

    Args:
        adata_sim (anndata.AnnData)
            Simulated AnnData object.

        state_key (str)
            Key in adata_sim.obs corresponding to annotated states.

        key_added (str)
            Key added to adata.obs to indicate the fate of each trajectory.
            **Default**: "fate"

        time_key (str)
            Key in adata_sim.obs corresponding to time.

        sim_key (str)
            Key in adata_sim.obs corresponding to each individual simulation.
            
        silent (bool)
            Description.

    Returns:
        None
    """

    fate_annot = CellFateAnnotation(time_key=time_key, sim_key=sim_key, silent=silent)
    fate_annot(adata_sim=adata_sim, state_key=state_key, key_added=key_added)
